#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <ctype.h>
#include <unistd.h>

#ifdef _DCC
# include <dos/stdio.h>
# include <dos/dos.h>
# include <exec/exec.h>
# include <clib/dos_protos.h>
# include <clib/exec_protos.h>
# define NOT_FOUND 6 
# define QFAX_PHONE "s:qfax.phone"
#else
#ifndef __linux__
# define Open fopen
#else
# define Open lopen
#endif
# define MODE_OLDFILE "r"
# define MODE_NEWFILE "w"
# define Close fclose
# define FGets(p1, p2, p3) fgets(p2, p3, p1)
# define PutStr(p1) fputs(p1, stderr)
# define BPTR FILE * 
# define Printf printf
# ifdef QDOS
#  define NOT_FOUND (-7)
#  define QFAX_PHONE "qfax_phone"
#  include <memory.h>
# else
#  define NOT_FOUND 6
#  define QFAX_PHONE "~/.qfax.phone"
# endif

#endif

#include "qfaxparse.h"

static FAXDATA defs = {
    NULL,
    {0},
#ifdef _DCC
    "serial.device",
    {0},
    {0},
    "temp:qfax",
    "temp:qfax/fax.log",
    "QFAX System",
    "s:qfax.font",
    "t:",
#elif defined (QDOS)
    "ser2hr",
    {0},
    {0},
    "ram1_fax_",
    "ram2_fax_log",
    "QFAX system",
    "flp1_qfax_font",
    "ram2_",
#else
    "/dev/cua0",
    {0},
    {0},
    "/usr/local/faxin/",
    "/usr/local/faxin/fax.log",
    "QFAX system",
# ifdef __VMS
    "sys$login:qfax.font",
# else
    ".qfax.font",
# endif
    "",
#endif
    {0},
    19200,
    -1,
    10,
    1,
    5,
    0,
    1,
    0,
    1,
    0x12,
    0,   /* devu */
    0,   /* width */
    2,   /* page */
    0,   /* wg3 */
    0,   /* pg3  */
    0,   /* paper */
    {0}, /* unlimited */
    0,   /* nodata */
    3,   /* ppr */
    160, /* lmargin */
    "",
    {{0},{0},{0},{0}},
    "D",
    "Z",
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
#ifdef QDOS
    {0},
#endif
    "",
    "fax",
    -1,     /* rmodem */
    0,      /* elsb */
    1,      /* poll */
    {0},
    0,
#ifdef QDOS
    {0},
#endif
    10,
    2,
    2,
    0,
    {0},
    {QFAX_PHONE},
#ifdef QDOS
    "ram2_script_log",
    {0},
    {0},
    4,
#endif
    {0},
    0,
   NULL,      /* prview */
   {0},      /* hangup */
   {0}       /* Font dir */

};

#ifdef __linux__
typedef unsigned char ubyte;
#endif

static XPL xpt[] = {
   {  0, 0, 1, "hp-pcl"},
   {  1, 0, 1, "eps-9"},
   {  2, 0, 1, "eps-24"},
   {  3, 0, 0, "ps"},
   {  4, 0, 0, "ps-lpr"},
   {100, 1, 0, "pbm"},
   {101, 1, 0, "pic"},
   {102, 1, 0, "pic-scaled"},
   {103, 0, 0, "gif"},
   {104, 1, 0, "gif-paged"},
   {-1,  0, 0, "???"}
};

XPL * XLateXPL(short t, char *n)
{
    XPL *x;
    for(x = xpt; x->type != -1; x++)
    {
	if(n)
	{
	    if(strncmp(n, x->name, strlen(n)) == 0)
	    {
		break;
	    } 
	}
	else if(t == x->type)
	{
	    break;
	}
    }
    return (x->type != -1) ? x : NULL;
}  

static char *strim (char *s)
{
    char *p;
    short n;

    n = strlen(s);
    p = s + n;
    while(*--p <= ' ' && n--)
    {
        *p = '\0';
    }
    return s;
}

static short AlMatch(char *s, char **p2)
{
    static char *opts[] = {"TSI","NAME", "ALIAS","DCODE","2D", "FULL", NULL};
    char **p;
    short n = 0;
    short sl;
    for(p = opts; *p; p++)
    {
        *p2 = NULL;
        if(strnicmp(s, *p, (sl = strlen(*p))) == 0)
        {
            {
                (*p2) = s + sl + strspn(s + sl, "= \t");
                break;
            }
        }
        n++;
    }
    return n;
}
#ifdef QDOS
static void  pSFlag (PFAXDATA p, char *s, int nx)
{
    strlwr(s);
    if(strchr(s, 's'))
    {
        p->sflag |= 1;
    }
    if(strchr(s, 'r'))
    {
        p->sflag |= 2;
    }
}
#endif
static void  pprview (PFAXDATA p, char *s, int nx)
{
    PRVIEW *pn;
    char *p0, *p1;
    short n,done;
    
    pn = (PRVIEW *) calloc(sizeof(PRVIEW), 1);
    if(pn)
    {
        for(n = done = 0, p0 = s; !done; n++)
        {
	    p1 = strchr(p0, ',');
	    if(p1)
	    {
		*p1 = '\0';
	    }
	    else done = 1;
	    
            p0 += strspn (p0, " \t");
	    
	    switch(n)
            {
	      case 0:
		{
		    XPL *x;
		    if(isdigit(*p0))
		    {
			x = XLateXPL(strtol(p0, NULL, 0), NULL);
		    }
		    else
		    {
			x = XLateXPL(0, p0);
		    }
		    if(x)
		    {
			pn->value = x->type;
		    }
		    else
		    {
			free(pn);
			return;
		    }
		}
		break;
              case 1:
		if(*p0)
		    pn->program = strdup(p0);
		break;
	      case 2:
		if(*p0)
		    pn->file = strdup(p0);
		break;
            }
	    if(p1) p0 = p1 + 1;
        }
	pn->next = p->prview;
	p->prview = pn;
    }
}

static void  pAlias (PFAXDATA p, char *s, int nx)
{
    char *p0, *p1,*p2;
    short n;
    ALIAS *pn;

    pn = (ALIAS *) calloc(sizeof(ALIAS), 1);

    if(pn)
    {
	pn->flag = 0;
        for(p0 = s; p1 = strtok(p0, ","); p0 = NULL)
        {
            p1 += strspn (p1, " \t");
            {
                n = AlMatch(p1, &p2);                            
                switch(n)
                {
                    case 0:
                    strncpy (pn->phone, p2, 39);
                    break;
                    case 1:
                    strncpy (pn->alias, p2, 39);
                    break;
                    case 2:
                    strncpy (pn->abbrev, p2, 15);
                    break;                
                    case 3:
                    strncpy (pn->dcode, p2, 39);
                    pn->flag &= ~0x8000;
                    break;
                    case 4:
                    pn->flag |= 1;
                    break;
                    case 5:
                    strncpy (pn->dcode, p2, 39);
                    pn->flag |= 0x8000;
                    break;
                }
            }
        }
	pn->next = p->alias;
	p->alias = pn;
    }
}

static void  pPhBook (PFAXDATA p, char *s, int nx)
{
    BPTR fd;
    char buf[256];
    
    if(fd = Open (s, MODE_OLDFILE))
    {
        while(FGets(fd, buf, sizeof(buf)))
        {
            char *pe;
            if(pe = strchr(buf, '\n'))
            {
                *pe = 0;
            }
            pAlias(p, buf, 0);
        }
        Close(fd);
    }
}
void FreeAlias(PFAXDATA p)
{
    ALIAS *pl, *dead;

    for(pl = p->alias; pl;)
    {
        dead = pl;
        pl = pl->next;
        free(dead);
    }
    p->alias = NULL;
}
static void  pString (char *p, char *s, int n)
{
    if(*s)
    {
        strncpy(p, s, n);
    }
}

static void  pModClr (char *p, char *s, int n)
{
    *p = 0;
}
static void  pLsbSet (FAXDATA *p, char *s, int n)
{
    p->elsb[0] = n + 2;
}
static void  pModSet(char *p, char *s, int n)
{
    if(s && *s)
    {
        strncpy(p, s, n);
    }
    else
    {
        *p = 0xff;
    }
}

static void  pRes(PFAXDATA p, char *s, int n)
{
    char rv = 0;
    if(stricmp(s, "high") == 0)
    {
        rv = 1;
    }
    else
    {
        if(atoi(s))
        {
            rv = 1;
        }
    }
    
    p->res = rv;
}

static void  pInt0 (int *p, char *s, int n)
{
    int rv = 0;

    rv = atoi(s);
    if(rv)
    {
        *p = rv;
    }
}
static void  pLMargn (short *p, char *s, int n)
{
    double f;
    char *q;
    short np = 0, unit = 0;

    f = strtod(s, &q);
    if(q && q != s)
    {
        if(strstr(q, "in"))
        {
            unit = 1;
        }
        else if(strstr(q, "mm"))
        {
            unit = 2;
        }
        switch(unit)
        {
        case 0:
            np = f;
            break;
        case 1:
            np = (f * 196.0+.5);
            break;
        case 2:
            np = (f * 7.7 +0.5);
            break;
        }
        *p = np;
    }
}

static void  pUnit (PFAXDATA p, char *s, int n)
{
    int rv = 0;

    rv = atoi(s);
    if(rv)
    {
        p->devunit = rv;
    }
}

static void  pPageWL(char * p, char *s, int n)
{
    char rv = 0;

    rv = atoi(s);
    if(rv >= 0 && rv < 3)
    {
        *p = rv;
    }
}

static void  pShort (short *p, char *s, int n)
{
    short rv = 0;

    rv = atoi(s);
    if(rv >= 0 && rv < 32767)
    {
        *p = rv;
    }
}

static void  pPPR (PFAXDATA p, char *s, int n)
{
    short rv = 0;

    rv = atoi(s);
    if(rv >= 0 && rv < 6)
    {
        p->ppr = rv;
    }
}

static void  pModem(char * p, char *s, int n)
{
    char rv = 0;

    rv = atoi(s);
    if(rv >= 0 && rv < 6)
    {
        *p = rv;
    }
}

static void  pCode(PFAXDATA p, char *s, int n)
{
    char rv = 0;
    
    if(atoi(s))
    {
        rv = 1;
    }
    p->code = rv;
}

static void  pSend (PFAXDATA p, char *s, int n)
{
    if(stricmp(s, "lsb") == 0)
    {
        p->sendrev = 1;
    }
}
#if 0
static void  pRecv (PFAXDATA p, char *s, int n)
{
    if(stricmp(s, "lsb") == 0)
    {
        p->recvrev = 1;
    }
}
#endif
static void  pWait (PFAXDATA p, char *s, int n)
{
    if(stricmp(s, "no") == 0 || stricmp(s, "off") == 0)
    {
        p->wait2send = 0;
    }
}

static void  pStart (PFAXDATA p, char *s, int n)
{
    if(stricmp(s, "xon") == 0)
    {
        p->startchar = 0x11;
    }
    if(stricmp(s, "none") == 0)
    {
        p->startchar = 0;
    }
}

static void  pPSize (PFAXDATA p, char *s, int n)
{

    if(!stricmp(s, "letter") || !stricmp(s, "ltr"))
    {
        p->paper = 1;
    }
    else if(!stricmp(s, "legal") || !stricmp(s, "lgl"))
    {
        p->paper = 2;
    }
    else if(!stricmp(s, "wide"))
    {
        p->paper = 4;
    }    
}

static void  pMString(char **p, char *s, int n)
{
    *p = strdup(s); 
}
#ifndef _DCC
# define VF (void(*)(void *,void *,int))
#else
#define VF
#endif

static PARSELIST pl[] = {
    {"ID",      VF pString,    offsetof(FAXDATA, LocalId),     63},
    {"SERD",    VF pString,    offsetof(FAXDATA, serdev),      31},
    {"INI",     VF pString,    offsetof(FAXDATA, init),        63},
    {"PREINI",  VF pString,    offsetof(FAXDATA, preinit),     63},
    {"IND",     VF pString,    offsetof(FAXDATA, in_dir),      63},
    {"INF",     VF pString,    offsetof(FAXDATA, info_file),   63},
    {"SIG",     VF pString,    offsetof(FAXDATA, Signature),   63},
    {"FONT",    VF pString,    offsetof(FAXDATA, Font),        39},
    {"TEMP",    VF pString,    offsetof(FAXDATA, Temp),        31},
    {"VIA",     VF pString,    offsetof(FAXDATA, via),         31},
    {"SUFFIX",  VF pString,    offsetof(FAXDATA, suffix),      15},
#ifdef QDOS
    {"CHILDLOG",VF pString,    offsetof(FAXDATA, nul),         39},
    {"SCRIPTL", VF pString,    offsetof(FAXDATA, ScriptLog),   39},
    {"STARTS",  VF pString,    offsetof(FAXDATA, StartScript), 39},
    {"STOPS",   VF pString,    offsetof(FAXDATA, StopScript),  39},
    {"RSFLAG",  VF pSFlag},
#endif
    {"DATA",    VF pString,    offsetof(FAXDATA, data),        63},
    {"SERB",    VF pInt0,      offsetof(FAXDATA, baud)},
    {"EXIT",    VF pInt0,      offsetof(FAXDATA, ebaud)},
    {"LEFTMAR", VF pLMargn,    offsetof(FAXDATA, lmargin)},
    {"WIDTH_S", VF pPageWL,    offsetof(FAXDATA, width_g3)},
    {"WIDTH_R", VF pPageWL,    offsetof(FAXDATA, width)},
    {"PAGE_S",  VF pPageWL,    offsetof(FAXDATA, page_g3)},
    {"PAGE_R",  VF pPageWL,    offsetof(FAXDATA, page)},
    {"DELAY",   VF pShort,     offsetof(FAXDATA, delay)},
    {"GDELAY",  VF pShort,     offsetof(FAXDATA, gdelay)},
    {"CHARDEL", VF pShort,     offsetof(FAXDATA, iwait)},
    {"RCLASS",  VF pShort,     offsetof(FAXDATA, rclass)},
    {"SCLASS",  VF pShort,     offsetof(FAXDATA, sclass)},
    {"FAA",     VF pShort,     offsetof(FAXDATA, faa)},
    {"CHARDEL", VF pShort,     offsetof(FAXDATA, iwait)},
    {"PPR",     VF pPPR},
    {"PAPERS",  VF pPSize},
    {"RES",      VF pRes},
    {"MOD",     VF pModem,     offsetof(FAXDATA, vmodem)},
    {"RMOD",    VF pModem,     offsetof(FAXDATA, rmodem)},
    {"COD",     VF pCode},
    {"SEND",    VF pSend},
#if 0
    {"RECV",    VF pRecv},
#endif
    {"WAIT",    VF pWait},
    {"STARTC",  VF pStart},
    {"UNIT",    VF pUnit},
    {"PHONEBOOK", VF pPhBook},
    {"PRINT",   VF pString,    offsetof(FAXDATA, pcmd),        63},
    {"ATZ",     VF pString,    offsetof(FAXDATA, inicmd),      15},
    {"DIAL",    VF pString,    offsetof(FAXDATA, dialcmd),     15},
#ifdef QDOS
    {"MONO",    VF pModSet,    offsetof(FAXDATA, mono),        1,1},
#endif
    {"SAVET",   VF pString,    offsetof(FAXDATA, arrivals),    39},
    {"XDIAL",   VF pString,    offsetof(FAXDATA, indial),      39},
    {"UNLIM",   VF pModSet,    offsetof(FAXDATA, unlimited),   1,1},
    {"MSB",     VF pLsbSet,    0,   0,1},
    {"LSB",     VF pLsbSet,    0,   1,1},
    {"MDM1",    VF pModSet,    offsetof(FAXDATA, ini),         64,1},
    {"MDM2",    VF pModSet,    offsetof(FAXDATA, ini) +  64,   64,1},
    {"MDM3",    VF pModSet,    offsetof(FAXDATA, ini) + 128,   64,1},
    {"MDM4",    VF pModSet,    offsetof(FAXDATA, ini) + 192,   64,1},
    {"QFAXINI", VF pMString,   offsetof(FAXDATA, qfaxini)},
    {"QFV2INI", VF pMString,   offsetof(FAXDATA, qfv2ini)},
    {"QFVINI",  VF pMString,   offsetof(FAXDATA, qfv2ini)},
    {"T2G3INI", VF pMString,   offsetof(FAXDATA, t2g3ini)},
    {"PFAXINI", VF pMString,   offsetof(FAXDATA, pfaxini)},
    {"MONTHS",  VF pMString,   offsetof(FAXDATA, mstr)},
    {"DAYS",    VF pMString,   offsetof(FAXDATA, dstr)},
    {"NOPOLL",  VF pModClr,    offsetof(FAXDATA, poll),          1,1},
    {"APRES",   VF pString,    offsetof(FAXDATA, apres),          31},
    {"GFLAG",   VF pShort,     offsetof(FAXDATA, tg3)},
    {"PRVIEW",  VF pprview},
    {"HANGUP",  VF pModSet,    offsetof(FAXDATA, hangup),        1,1},
    {"PATHFO",  VF pString,    offsetof(FAXDATA, fontdir),        63},
   {NULL, NULL}
};

void FaxParams (PFAXDATA p)
{
    BPTR fp;
    char *pf;
    
#ifdef _DCC
    char pname[64];

    if(GetVar("QFAX_DATA", (pf = pname), sizeof(pname), 0) <= 0)
    {
        pf = QFAX_FILE;
    }
#else
    if((pf = getenv("QFAX_DATA")) == NULL)
    {
#ifdef QDOS
        pf = qfd.name;
#else
	char pname[255];
	if(pf = getenv("HOME"))
	{
	    short n;
	    
	    strcpy(pname, pf);
	    n = strlen(pname);
	    if(*(pname + n-1) != '/')
	    {
		*(pname + n) = '/';
		*(pname + n+1) = '\0';
	    }
	    
	}
	else
	{
	    *pname = 0;
	}
	strcat(pname, QFAX_FILE);
	pf = pname;
#endif
    }
#endif

    *p = defs;

    if((fp = Open(pf, MODE_OLDFILE)) != NULL)
    {
        char text [128];
	
        while(FGets(fp, text, sizeof(text)))
        {
            char *s;
            PPARSELIST ppl;
            char *ptr;

            ptr = (char *)p;
            strim(text);

            for(ppl = pl; ppl->id; ppl++)
            {
                if(strnicmp(text, ppl->id, strlen(ppl->id)) == 0)
                {
                    int l;
                    if((s = strchr(text, '=')))
                    {
                        s++;
                        l = strspn(s, " \t");
                    }
                    else if(ppl->flag)
                    {
                        s = text+strlen(text);
                        l = 0;
                    }
                    if(s)
                    {
                        (ppl->func)((ptr + ppl->offset), s + l, ppl->mval);
                        break;
                    }
                }
            }
        }
        Close(fp);
    }
    else
    {
	perror("failed to open");
	
        p->nodata = 1;
    }
}

void FrigArgs(int *pac, char ***pav, char *p)
{
    int n;
    char *p1,*p2,*p3;
    char **xav, **x, **sav;

    sav = *pav;
    p3 = strdup(p);
    for(n = 0, p1 = p; p2 = strtok(p1, " \t"); p1 = NULL, n++)
        ;

    n += ((*pac) + 1);
    xav = x = malloc(n * sizeof(char *));
    *x++ = *(*pav)++;
    for(p1 = p3; p2 = strtok(p1, " \t"); p1 = NULL)
    {
        *x++ = strdup(p2);
    }
    while(--(*pac))
    {
        *x++ = *(*pav)++;
    }
    *x = NULL;
    *pac = n-1;
#ifndef __linux__
    if(sav)
    {
        free(sav);
    }
#endif
    if(p3)
    {
        free(p3);
    }
    *pav = xav;
}

#ifdef DEBUG

void Explain(char *p, char *q)
{
    if(*q == 0)
    {
        q = "Use QFAX default";
    }
    else if (*q == -1)
    {
        q = "Ignore completely";
    }

    Printf("%s%s\n", p, q);        
    
}

int main(int ac, char **av)
{
    FAXDATA ff;
    
    FaxParams(&ff);

    Printf ("Local ID        : %s\n", ff.LocalId);
    Printf ("Serial          : %s\n", ff.serdev);
    Printf ("Modem Init      : %s\n", ff.init);
    Printf ("In bound        : %s\n", ff.in_dir);
    Printf ("Info File       : %s\n", ff.info_file);
    Printf ("Signature       : %s\n", ff.Signature);
    Printf ("Font file       : %s\n", ff.Font);
    Printf ("Temp            : %s\n", ff.Temp);
    Printf ("Via             : %s\n", ff.via);
    Printf ("Baud            : %ld\n", (long) ff.baud);
    Printf ("EBaud           : %ld\n", (long) ff.ebaud);
    Printf ("Delay           : %ld\n", (long) ff.delay);
    Printf ("Res             : %ld\n", (long) ff.res);
    Printf ("Modem level     : %ld\n", (long) ff.vmodem);
    Printf ("Coding          : %ld\n", (long) ff.code);
    Printf ("Send order      : %ld %lcsb\n", (long) ff.sendrev, (long) (ff.sendrev) ? 'l': 'm');
    Printf ("Wait2Send       : %ld\n", (long) ff.wait2send);
    Printf ("StartChar       : %lx hex\n", (long) ff.startchar);
    Printf ("Dev Unit        : %ld\n", (long) ff.devunit);
    Printf ("Print Command   : %s\n", ff.pcmd);
    Printf ("Modem Reset     : %s\n", ff.inicmd);
    Printf ("Pre Dial Command: %s\n", ff.dialcmd);
    Explain("Modem Setup 1   : ", ff.ini[0]);
    Explain("Modem Setup 2   : ", ff.ini[1]);
    Explain("Modem Setup 3   : ", ff.ini[2]);
    Explain("Modem Setup 4   : ", ff.ini[3]);
    Printf ("RModem level    : %ld\n", (long) ff.rmodem);
    Printf ("qfax ini        : %s\n", (ff.qfaxini) ? ff.qfaxini : "");
    Printf ("qfv2 ini        : %s\n", (ff.qfv2ini) ? ff.qfv2ini : "");
    Printf ("t2g3 ini        : %s\n", (ff.t2g3ini) ? ff.t2g3ini : "");
    Printf ("pfax ini        : %s\n", (ff.pfaxini) ? ff.pfaxini : "");

    if(ff.prview)
    {
	PRVIEW *p;
	for(p = ff.prview; p; p = p->next)
	{
	    printf("PRVIEW: %d %s %s\n",
		   p->value,
		   (p->program) ? p->program : "(none)",
		   (p->file) ? p->file : "(none)");
	}
    }
    

    if(ff.alias)
    {
        BPTR fd = 0;
        ALIAS *pl;
#ifdef _DCC
        if(ac == 2)
        {
            fd = Open(QFAX_PHONE, MODE_NEWFILE);
        }
#endif
        for(pl = ff.alias; pl; pl = pl->next)
        {
            Printf("Alias: %s %s %s %lx (%s)\n",
            pl->alias, pl->dcode, pl->phone, pl->flag, pl->abbrev);
#ifdef _DCC
            if(fd)
            {
                if(*pl->alias)
                {
                    FPrintf(fd,"NAME = %s, ", pl->alias);
                }
                
                if(*pl->abbrev)
                {
                    FPrintf(fd,"ALIAS = %s, ", pl->abbrev);
                }
                if(*pl->phone)
                {
                    FPrintf(fd,"TSI = %s, ", pl->phone);
                }
                if(*pl->dcode)
                {
                    if(pl->flag & 0x8000)
                    {
                        FPrintf(fd,"FULL = %s\n", pl->dcode);
                    }
                    else
                    {
                        FPrintf(fd,"DCODE = %s\n", pl->dcode);
                    }
		} 
            }
#endif    
	}
        FreeAlias(&ff);
#ifdef _DCC
        if(fd)
        {
            Close(fd);
        }
#endif
    }
    return 0;
}
#endif
