#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#ifdef QDOS
#define bzero(a,b) memset(a,0,b)
#endif
#include "qheader.h"
#include "g3encoder.h"
#include "alphabet.h"

static CODERSTATE *code = NULL;

static int get_nce (int bp)
{
    short cu_co;

    cu_co = ((code->refline[bp >> 3] & code->mbits[bp & 0x07]) != 0);

    code->lce_col = !cu_co;

    if (cu_co)
    {
	while (((code->refline[bp >> 3] & code->mbits[bp & 0x07]))
                     && (bp < code->pix))
	{
	    bp++;
	}
    }
    else
    {
	while ((!(code->refline[bp >> 3] & code->mbits[bp & 0x07]))
                     && (bp < code->pix))
	{
	    bp++;
	}
    }
    return bp;
}

static UCHAR *setbits;
static UCHAR bits_lsb[] = {1,2,4,8,0x10, 0x20, 0x40, 0x80};

static void IncBitNum(void)
{
	code->bitnr++;
        if(code->bitnr == OUTBUFSIZE * 8)
        {
            FaxFlushBuf(OUTBUFSIZE);
        }
}

static void enter_bits (char *str)
{

    while (*str)
    {
	if (*str == '1')
	{
	    code->outbuf[code->bitnr >> 3] |= setbits[code->bitnr & 0x07L];
	}
	str++;
        IncBitNum();
    }
}

int FaxFlushBuf(short ws)
{
    int nw = 0;

    if(ws)
    {
        nw = write (code->hd, code->outbuf, ws);
        bzero ((char *)code->outbuf, (unsigned int) OUTBUFSIZE);
    }
    code->bitnr = 0;
    code->outptr = code->outbuf;
    return nw;
}

static short align = 0;

static int CoderEOL (void)
{
    if(code->gflag == 1 || align)
    {
        short ebit;
        ebit = ((code->bitnr + 19 + code->pack) & ~7) - (code->bitnr + 12 + code->pack);
        while(ebit--)
        {
            IncBitNum();
        }
    }
    if(align)
    {
        align = 0;
        code->line_2d= code->cnt_2d = 1;
    }

    enter_bits (EOL);
    if (code->pack)
    {
	if (code->line_2d)
	{
	    if ((--code->cnt_2d) == 0)
	    {
		code->line_2d = 0;
	    }
	}
	else
	{
            code->cnt_2d = code->cnt_ref;
	    code->line_2d = 1;
	}
	if (code->line_2d)
	{
	    enter_bits ("0");
	}
	else
	{
	    enter_bits ("1");
	}
    }
    code->scan++;
    return 0;
}

int FaxCodeAlign (UCHAR * pixptr)
{
    align = 1;
    return FaxCodeLine (pixptr);
}
void FaxCloseFile (void)
{
    FaxFlushBuf(((code->bitnr+7) >> 3));
    close (code->hd);
    code->hd = 0;
}

static int enter_white (int anz)
{
    if (anz > 63)
    {
	enter_bits ((char *) mu_code_white[(anz >> 6) - 1]);
	anz &= 0x3F;
    }
    enter_bits ((char *) white_codes[anz]);

    return 1;
}

static int enter_black (int anz)
{

    if (anz > 63)
    {
	enter_bits (mu_code_black[(anz >> 6) - 1]);
	anz &= 0x3F;
    }
    enter_bits ((char *) black_codes[anz]);

    return 0;
}

static int coder_2d (UCHAR * pixptr)
{
    int i, a1, dist, cc_cnt, j, u;
    int cur_col = 0, old_cc;

    i = 0;
    while (i < code->pix)
    {
	code->b1 = get_nce (i);	/* next changing element on refline */
	if (code->lce_col == cur_col)
	{
	    code->b1 = get_nce (code->b1);
	}

	a1 = i;

	if (cur_col)
	{
	    while (((pixptr[a1 >> 3] & code->mbits[a1 & 0x07])) &&
                                             (a1 < code->pix))
	    {
		a1++;
	    }
	}
	else
	{
	    while ((!(pixptr[a1 >> 3] & code->mbits[a1 & 0x07])) &&
                                            (a1 < code->pix))
	    {
		a1++;
	    }
	}

	if (a1 == code->b1)
	{
	    enter_bits (VERTICAL);
	    cur_col = !cur_col;
	    i = a1;
	}
	else
	{
	    code->b2 = get_nce (code->b1);
	    if (code->b2 < a1)
	    {			/* pass mode */
		i = code->b2;
		enter_bits (PASSMODE);
	    }
	    else if ((dist = (abs (code->b1 - a1))) <= 3)
	    {
		if ((code->b1 - a1) < 0)
		{
		    enter_bits ((char *) vertical_r[dist - 1]);
		}
		else
		{
		    enter_bits ((char *) vertical_l[dist - 1]);
		}
		cur_col = !cur_col;
		i = a1;
	    }
	    else
	    {			/* horizontal mode */

		enter_bits (HORMODE);
		cc_cnt = 0;

		for (u = 0; u < 2; u++)
		{
		    old_cc = cur_col;
		    for (j = i; (j < code->pix); j++)
		    {
			if ((pixptr[j >> 3]) & (code->mbits[j & 0x07]))
			{
			    if (cur_col)
				cc_cnt++;
			    else
			    {
				cur_col = enter_white (cc_cnt);
				cc_cnt = 0;
			    }
			}
			else
			{
			    if (cur_col)
			    {
				cur_col = enter_black (cc_cnt);
				cc_cnt = 0;
			    }
			    else
				cc_cnt++;
			}
			if (old_cc != cur_col)
			{
			    break;
			}
		    }
		    i = j;
		}		/* end coding a0->a1, a1->a2 */
		if (cc_cnt)
		{
		    if (cur_col)
			cur_col = enter_black (cc_cnt);
		    else
			cur_col = enter_white (cc_cnt);
		}
	    }
	}
    }
    return 0;
}
static void ClearCoder(void)
{
    if(code)
    {
        bzero ((char *)code->outbuf, (unsigned int) OUTBUFSIZE);
        code->outptr = code->outbuf;
        code->scan = code->bytenr = 0;
        code->bitnr = 0;
        code->line_2d = 0;
        code->cnt_ref = (code->res) ? K_CONST_H : K_CONST_L;
        code->cnt_2d = code->cnt_ref;
    }
}
void FaxBeginPage (void)
{
    ClearCoder();
/*    if(code->gflag) */
    {
        code->bitnr=4 - code->pack;
    }
    enter_bits (EOL);
    if (code->pack)
    {
	enter_bits ("1");
	code->line_2d = 0;
    }
}

int FaxCodeLine (UCHAR * pixptr)
{
    int i, curr_col = 0, cc_count = 0;

    if (code->pack && code->line_2d)
    {
	coder_2d (pixptr);
    }
    else
    {
	for (i = 0; i < code->pix; i++)
	{
	    if ((pixptr[i >> 3]) & (code->mbits[i & 0x07]))
	    {
		if (curr_col)
		    cc_count++;
		else
		{
		    curr_col = enter_white (cc_count);
		    cc_count = 1;
		}
	    }
	    else
	    {
		if (curr_col)
		{
		    curr_col = enter_black (cc_count);
		    cc_count = 1;
		}
		else
		    cc_count++;
	    }
	}

	if (cc_count)
	{
	    if (curr_col)
		enter_black (cc_count);
	    else
		enter_white (cc_count);
	}
    }

    if(!code->gflag)
    {
        while (code->bitnr & 0x07)
        {
            IncBitNum();
        }
        for(i =0; i < (12-code->pack); i++)
        {
            IncBitNum();
        }
    }    
    if (code->pack)
    {
	memcpy (code->refline, pixptr, (unsigned int) ((code->pix) >> 3));
    }
    i = CoderEOL ();
    return i;
}

void FaxCodeEOP (void)
{
    short i;

    /*
     * Only do FIVE for RTC as we've just done one
     */

    if(code->scan > code->maxlines)
    {
       code->maxlines = code->scan;
    }
    for (i = 0; i < 5; i++)
    {
	CoderEOL ();
    }

    if(code->bitnr)
    {
        FaxFlushBuf(((7+code->bitnr) >> 3));
    }
}

CODERSTATE * FaxInitCoder (short lsb)
{
    short i,k;
    if((code = (CODERSTATE *) malloc(sizeof(CODERSTATE))))
    {
        bzero((char *) code, sizeof(CODERSTATE));
        code->lsb = lsb;
        if(lsb)
        {
            setbits = bits_lsb;
        }
        else
        {
            setbits = code->mbits;
        }
        k = 0x80;
        for(i = 0; i < 8; i++)
        {
            code->mbits[i] = k;
            k = k >> 1;
        }
    }
#ifdef __linux__
    code->gflag = 1;
#endif

    return code;
}

int FaxOpenFile(char *outfn, short resolution, short packmode, short pix)
{
    int rv = -1;

    if(code)
    {
        code->res = resolution;
        code->pack = packmode;
        code->pix = pix;
    
        if(outfn)
        {
            if((code->hd = open (outfn, O_WRONLY | O_TRUNC | O_CREAT)) > 0)
            {
                rv = code->hd;
            }
        }
        ClearCoder();
    }
    return rv;
}

void FaxFreeCoder(CODERSTATE *dummy)   /* We know what we really want :-) */
{
    if(code->hd)
    {
        FaxCloseFile();
    }
    free(code);
    code = NULL;
}

/* Write QFAX Header.
   hd :  File Number
   Page: Number of pages in file
 */
#ifdef __linux__
unsigned short SH(unsigned short w)
{
       union {
               unsigned short w;
               char c[2];
       } u,v;
       u.w = w;
       v.c[0] = u.c[1];
       v.c[1] = u.c[0];
       return v.w;
}

unsigned long LG(unsigned long l)
{
       union {
               unsigned long l;
               char c[4];
       } u,v;

       u.l = l;

       v.c[0] = u.c[3];
       v.c[1] = u.c[2];
       v.c[2] = u.c[1];
       v.c[3] = u.c[0];
       return v.l;
}
#else
#define SH(x) (x)
#define LG(x) (x)
#endif

void FaxHeader (short page, short width, short length, char *tsi, long offset)
{
    QFAXHEADER h;
    if(code->hd > 0)
    {
        lseek(code->hd, 0, 0);
        bzero ((char *) &h, QFAXHEADLEN);   /* place keeper */
        if (page)
        {
	    h.title = LG(QFAXTITLE);
	    h.version = LG(QFAXVERS);
	    h.width = SH(width);
	    h.resolution = code->res;
	    h.coding = code->pack;
	    h.page = SH(length);
	    h.pagemax = SH(page);
	    if(tsi)
	    {
                strcpy (h.tsi, tsi); /* Local TSI, can be blank */
            }
	    h.stamp = LG(time (NULL));
	    h.hint = LG(code->maxlines);     /* Max lines in a page this fax */
	    h.offset = LG(offset);
	    h.lsb = code->lsb;
        }
        write (code->hd, &h, QFAXHEADLEN);
    }
}

